#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <assert.h>

#include "treap.h"

static void test_treap_alloc() {
	struct treap *t = treap_alloc();
	assert(t);
	assert(t->n == 0);
	treap_dealloc(t);
}


static void test_treap_insert() {
	struct treap *t = treap_alloc();
	const int max_items = 10000;
	int i = max_items;
	clock_t start, stop;

	assert(t);
	t->cmp = treap_int_cmp;

	/* insert data to the treap */
	start = clock();
	while (i--) treap_insert(t, i);
	stop = clock();

	printf("-- %d inserts: elapsed time: %f\n",
			max_items,
			(double)(stop - start)/CLOCKS_PER_SEC);

	assert(t->n == max_items);
	assert(treap_min(t));
	assert(treap_max(t));

	assert(treap_min(t)->d == 0);
	assert(treap_max(t)->d == ( max_items - 1 ));

	treap_dealloc(t);
}


static void test_treap_lookup() {
	struct treap *t = treap_alloc();
	const int max_items = 10000;
	int i = max_items;
	clock_t start, stop;

	assert(t);
	t->cmp = treap_int_cmp;

	/* insert data to the treap */
	while (i--) treap_insert(t, i);

	/* finding a non existing item should fail */
	assert(treap_find(t, max_items + 1) == NULL);

	/* finding existing item should succeed */
	assert(treap_find(t, max_items - 1) != NULL);

	/* verify if all inserted items exist in the treap */
	i = max_items;
	start = clock();
	while (i--) {
		struct treap_node *n = treap_find(t, i);
		assert(n);
		assert(n->d == i);
	}
	stop = clock();

	printf("-- %d lookups: elapsed time: %f\n",
			max_items,
			(double)(stop - start)/CLOCKS_PER_SEC);

	treap_dealloc(t);
}


struct int_array {
	size_t max_size;
	size_t n;
	int data[];
};



static void _treap_array_push_visitor(struct treap_node *n, void *p, int l) {
	struct int_array *a = p;
	a->data[a->n++] = n->d;
}


static void test_treap_traverse() {
	struct treap *t = treap_alloc();
	const int max_items = 10000;
	int i = max_items;

	struct int_array *a =
		malloc(sizeof(struct int_array) + (sizeof(int) * max_items ));

	assert(a);
	assert(t);
	t->cmp = treap_int_cmp;

	a->max_size = max_items;
	a->n = 0;

	/* insert data to the treap */
	while (i--) treap_insert(t, i);

	/* verify the order (in order traversal - sorted order) */
	treap_traverse_in(t, _treap_array_push_visitor, a);
	for (i = 0; i < max_items; i++) {
		assert(a->data[i] == i);
	}

	free(a);
	treap_dealloc(t);
}


static void test_treap_delete() {
	struct treap *t = treap_alloc();
	const int max_items = 10000;
	const int deleted_items = 1000;
	int i = max_items;
	int *deleted = malloc(sizeof(int) * max_items);

	assert(t);
	assert(deleted);
	t->cmp = treap_int_cmp;

	/* insert data to the treap */
	while (i--) treap_insert(t, i);

	/* delete data from the treap */
	for (i = 0; i < deleted_items; i++) {
		struct treap_node *n = NULL;

		do {
			int v = rand() % max_items;
			deleted[i] = v;
			n = treap_find(t, v);
		} while (!n);

		treap_delete(t, n);
	}

	assert(t->n == (max_items - deleted_items));

	/* verify if all deleted items are really gone from the treap */
	for (i = 0; i < deleted_items; i++) {
		assert(treap_find(t, deleted[i]) == NULL);
	}

	/* verify if all not deleted items are still in the treap */
	for (i = 0; i < max_items; i++) {
		struct treap_node *n = treap_find(t, i);
		int is_deleted = 0;
		int j = 0;

		/* check if item has been deleted */
		for (j = 0; j < deleted_items; j++) {
			if (i == deleted[j]) {
				is_deleted = 1;
				break;
			}
		}

		if (n) {
			assert(0 == is_deleted);
		}
		else {
			assert(1 == is_deleted);
		}
	}

	free(deleted);
	treap_dealloc(t);
}


static void test_treap_bh() {
	struct treap *t = treap_alloc();
	const int max_items = 10000;
	int i = max_items;

	assert(t);
	t->cmp = treap_int_cmp;

	/* insert data to the treap */
	while (i--) treap_insert(t, i);

	printf("-- nodes: (%d), max-height: %d, balance: (%d)\n",
			t->n,
			treap_max_height_get(t),
			treap_balance_factor_get(t));

	/* delete one third */
	i = max_items/3;
	while (i--) treap_delete(t, treap_find(t, i));

	printf("-- nodes: (%d), max-height %d, balance: (%d)\n",
			t->n,
			treap_max_height_get(t),
			treap_balance_factor_get(t));

	/* add more data */
	i = max_items;
	while (i--) treap_insert(t, max_items + i);
	printf("-- nodes: (%d), max-height: %d, balance: (%d)\n",
			t->n,
			treap_max_height_get(t),
			treap_balance_factor_get(t));

	treap_dealloc(t);
}


void test_treap_modify() {
	struct treap *t = treap_alloc();
	t->cmp = treap_int_cmp;

	treap_insert(t, 123);
	treap_insert(t, 456);
	treap_insert(t, 55);
	treap_insert(t, 22);
	treap_insert(t, 65);
	treap_modify(t, 22, 33);

	assert(treap_find(t, 22) == NULL);
	assert(treap_find(t, 33) != NULL);

	treap_delete(t, t->root);
	assert(t->root != NULL);

	treap_dealloc(t);
}


static struct test_case {
	const char *tc_name;
	void (*tc_cb)(void);
} test_cases[] = {
	{ "treap_alloc test", test_treap_alloc },
	{ "treap_insert test", test_treap_insert },
	{ "treap_lookup test", test_treap_lookup },
	{ "treap traverse test", test_treap_traverse },
	{ "treap delete test", test_treap_delete },
	{ "treap balance & height test", test_treap_bh },
	{ "treap modify test", test_treap_modify },
	{ NULL, NULL },
};


int main(int argc, char *argv[]) {
	struct test_case *tc = test_cases;

	srand(time(NULL));

	while (tc->tc_name && tc->tc_cb) {
		printf("executing: %s ... \n", tc->tc_name);
		tc->tc_cb();
		tc++;
	}

	printf("testing finished\n");
	return 0;
}
