#include "treap.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

/* ========================================================================== */

static struct treap_node* _treap_node_alloc(treap_data_t d);
static int _treap_node_what_leaf(struct treap_node *p, struct treap_node *n);
static int _treap_node_is_leaf(struct treap_node *n);
static int _treap_node_leafs_count(struct treap_node *n);
static struct treap_node* _treap_node_parent_find(struct treap_node *h, struct treap_node *n);

static void _treap_rotate_left(struct treap *t, struct treap_node *n);
static void _treap_rotate_right(struct treap *t, struct treap_node *n);
static void _treap_traverse_in(struct treap_node *n, treap_visitor_t v, void *priv);
static void _treap_treapify_up(struct treap *t, struct treap_node *n);
static void _treap_treapify_down(struct treap *t, struct treap_node *n);

static int _treap_max_height_get(struct treap_node *n);

/* ========================================================================== */

struct treap* treap_alloc() {
	struct treap *t = malloc(sizeof(struct treap));
	memset(t, 0x00, sizeof(struct treap));
	return t;
}


void treap_dealloc(struct treap *t) {
	assert(t);

	while (t->n) {
		treap_delete(t, treap_min(t));
	}

	free(t);
}


void treap_insert(struct treap *t, treap_data_t d) {
	struct treap_node *node = _treap_node_alloc(d);
	struct treap_node **n = NULL;

	assert(t);
	assert(t->cmp);
	assert(node);

	n = &t->root;
	while(*n) {
		treap_leaf l = (1 == t->cmp(&((*n)->d), &d)) ? L_LEFT : L_RIGHT;
		n = &((*n)->l[l]);
	}

	*n = node;
	t->n++;

	_treap_treapify_up(t, node);
}


void treap_delete(struct treap *t, struct treap_node *n) {
	struct treap_node *p = NULL;
	treap_leaf l = L_MAX;

	assert(t);
	assert(n);

	if (_treap_node_leafs_count(n))
		_treap_treapify_down(t, n);

	p = treap_parent_find(t, n);
	if (p) {
		l = _treap_node_what_leaf(p, n);
		p->l[l] = NULL;
	}

	t->n--;
	free(n);
}


void treap_modify(struct treap *t, treap_data_t o, treap_data_t n) {
	struct treap_node *node = treap_find(t, o);

	if (!node)
		return;

	treap_delete(t, node);
	treap_insert(t, n);
}


struct treap_node* treap_find(struct treap *t, treap_data_t d) {
	struct treap_node *n = t->root;
	int r = 0;

	assert(t);
	assert(t->cmp);

	while (n && (r = t->cmp(&n->d, &d))) {
		treap_leaf l = (1 == r) ? L_LEFT : L_RIGHT;
		n = n->l[l];
	}

	return n;
}


struct treap_node* treap_min(struct treap *t) {
	struct treap_node *n = t->root;
	while (n->l[L_LEFT]) {
		n = n->l[L_LEFT];
	}

	return n;
}


struct treap_node* treap_max(struct treap *t) {
	struct treap_node *n = t->root;
	while (n->l[L_RIGHT]) {
		n = n->l[L_RIGHT];
	}
	return n;
}


void treap_traverse_in(struct treap *t, treap_visitor_t v, void *priv) {
/* #define USE_ITER */
#ifndef USE_ITER
	_treap_traverse_in(t->root, v, priv);
#else
	struct treap_node **stack = malloc(sizeof(struct treap_node *) * t->n);
	struct treap_node *node = t->root;
	size_t sp = 0;

	while (node || sp) {
		if (node) {
			stack[sp++] = node;
			node = node->l[L_LEFT];
		}
		else {
			node = stack[--sp];
			v(node, priv, 0);
			node = node->l[L_RIGHT];
		}
	}

	free(stack);
#endif
}


void treap_traverse_level(struct treap *t, treap_visitor_t v, void *priv) {
	const size_t qsize = ( t->n + 1 );
	struct treap_node **queue =
		malloc(sizeof(struct treap_node *) * qsize);
	int qh = 0;
	int qt = 0;

	int curr_lvl_cnt = 0;
	int next_lvl_cnt = 0;

	assert(queue);
	queue[qh++] = t->root;
	curr_lvl_cnt++;

#define QEMPTY(__qh, __qt) ((__qh) == (__qt))

	while (!QEMPTY(qh,qt)) {
		struct treap_node *node = queue[qt];
		qt = (qt + 1) % qsize;

		curr_lvl_cnt--;
		v(node, priv, curr_lvl_cnt);

		if (node->l[L_LEFT]) {
			queue[qh] = node->l[L_LEFT];
			qh = (qh + 1) % qsize;
			next_lvl_cnt++;
		}

		if (node->l[L_RIGHT]) {
			queue[qh] = node->l[L_RIGHT];
			qh = (qh + 1) % qsize;
			next_lvl_cnt++;
		}

		if (0 == curr_lvl_cnt) {
			/* printf("\n"); */
			curr_lvl_cnt = next_lvl_cnt;
			next_lvl_cnt = 0;
		}
	}

	free(queue);
}


int treap_max_height_get(struct treap *t) {
	_treap_max_height_get(t->root);
}


int treap_balance_factor_get(struct treap *t) {
	int l = _treap_max_height_get(t->root->l[L_LEFT]);
	int r = _treap_max_height_get(t->root->l[L_RIGHT]);

	return (r - l);
}


struct treap_node* treap_parent_find(struct treap *t, struct treap_node *n) {
	return _treap_node_parent_find(t->root, n);
}


/* ========================================================================== */

static int _treap_max_height_get(struct treap_node *n) {
	int hl = 0;
	int hr = 0;

	if (!n) return 0;
	if (n->l[L_LEFT]) hl = 1 + _treap_max_height_get(n->l[L_LEFT]);
	if (n->l[L_RIGHT]) hr = 1 + _treap_max_height_get(n->l[L_RIGHT]);

	return (hl > hr ? hl : hr);
}


static void _treap_rotate_left(struct treap *t, struct treap_node *n) {
	struct treap_node *r = n->l[L_RIGHT];
	struct treap_node *p = treap_parent_find(t, n);

	assert(r);

	if (p) {
		treap_leaf li = _treap_node_what_leaf(p, n);
		p->l[li] = r;
	}

	n->l[L_RIGHT] = r->l[L_LEFT];
	r->l[L_LEFT] = n;

	if (n == t->root) {
		t->root = r;
	}
}


static void _treap_rotate_right(struct treap *t, struct treap_node *n) {
	struct treap_node *l = n->l[L_LEFT];
	struct treap_node *p = treap_parent_find(t, n);

	assert(l);

	if (p) {
		treap_leaf li = _treap_node_what_leaf(p, n);
		p->l[li] = l;
	}

	n->l[L_LEFT] = l->l[L_RIGHT];
	l->l[L_RIGHT] = n;

	if (n == t->root) {
		t->root = l;
	}
}


static struct treap_node* _treap_node_alloc(treap_data_t d) {
	struct treap_node *n = malloc(sizeof(struct treap_node));
	assert(n);

	/* clear the memory */
	memset(n, 0x00, sizeof(struct treap_node));

	/* assign the data */
	n->d = d;

	/* assign random priority */
	n->p = rand();
	return n;
}


static void _treap_traverse_in(struct treap_node *n, treap_visitor_t v, void *priv) {
	if (n) {
		_treap_traverse_in(n->l[L_LEFT], v, priv);
		v(n, priv, 0);
		_treap_traverse_in(n->l[L_RIGHT], v, priv);
	}
}


static struct treap_node* _treap_node_parent_find(struct treap_node *h, struct treap_node *n) {
	struct treap_node *p = NULL;

	if (h) {
		if ((n == h->l[L_LEFT]) || (n == h->l[L_RIGHT])) {
			return h;
		}

		p = _treap_node_parent_find(h->l[L_LEFT], n);
		if (!p) p = _treap_node_parent_find(h->l[L_RIGHT], n);
	}

	return p;
}


static int _treap_node_what_leaf(struct treap_node *p, struct treap_node *n) {
	return (p->l[L_LEFT] == n ? L_LEFT : (p->l[L_RIGHT] ? L_RIGHT : -1));
}


static int _treap_node_is_leaf(struct treap_node *n) {
	return (!n->l[L_LEFT] && !n->l[L_RIGHT]);
}


static int _treap_node_leafs_count(struct treap_node *n) {
	int c = 0;
	treap_leaf l = L_MAX;
	while (l--) if (n->l[l]) c++;
	return c;
}


static void _treap_treapify_up(struct treap *t, struct treap_node *n) {
	struct treap_node *p = treap_parent_find(t, n);

	while (p && (p->p > n->p)) {
		treap_leaf l = _treap_node_what_leaf(p, n);
		assert(l != -1);

		if (L_LEFT == l) {
			_treap_rotate_right(t, p);
		}
		else {
			_treap_rotate_left(t, p);
		}

		p = treap_parent_find(t, n);
	}

	p = treap_parent_find(t, n);
	if (!p) {
		t->root = n;
	}
}


static void _treap_treapify_down(struct treap *t, struct treap_node *n) {
	int nc = 0;
	while ((nc = _treap_node_leafs_count(n))) {
		switch(nc) {
			case 1:
				if (n->l[L_LEFT])
					_treap_rotate_right(t, n);
				else
					_treap_rotate_left(t, n);
				break;

			case 2:
				if (n->l[L_RIGHT]->p > n->l[L_LEFT]->p)
					_treap_rotate_right(t, n);
				else
					_treap_rotate_left(t, n);
				break;
		} /* switch */

		if (n == t->root) {
			struct treap_node *p = treap_parent_find(t, n);
			assert(p);
			t->root = p;
		}
	} /* while */
}

/* ========================================================================== */
