#ifndef TREAP_H_IGVJQTB6
#define TREAP_H_IGVJQTB6

#include <stdlib.h>
#include "treap_data_type.h"


/**
 * @brief leaf enumeration
 */
typedef enum _treap_leaf {
	L_LEFT = 0,
	L_RIGHT,
	L_MAX,
} treap_leaf;


/**
 * @brief treap node declaration
 */
struct treap_node {
	struct treap_node *l[L_MAX];

	/* random priority */
	int p;

	/* the data */
	treap_data_t d;
};


/**
 * @brief comparison routine
 *
 * @param a value
 * @param b value
 *
 * @return a == b -> 0, a > b -> 1, a < b -> -1
 */
typedef int (*treap_cmp_t)(treap_data_t*,treap_data_t*);


/**
 * @brief the visitor function
 *
 * @param treap
 * @param private data
 * @param private number of leafs left to traverse (relevant for BFS only)
 *
 */
typedef void (*treap_visitor_t)(struct treap_node*, void*, int);


/**
 * @brief treap data type declaration
 */
struct treap {
	struct treap_node *root;
	treap_cmp_t cmp;
	size_t n;
};


/**
 * @brief allocates a treap
 *
 * @return a treap pointer on success or NULL on failure
 */
struct treap* treap_alloc(void);

/**
 * @brief de-allocates treap
 *
 * @param treap treap pointer
 */
void treap_dealloc(struct treap*);

/**
 * @brief inserts data into the treap
 *
 * @param t treap
 * @param d data to be places in the treap
 */
void treap_insert(struct treap *t, treap_data_t d);

/**
 * @brief delete a node from the treap
 *
 * @param t treap
 * @param n node to be deleted
 */
void treap_delete(struct treap *t, struct treap_node *n);

/**
 * @brief modify a data in the treap
 *
 * This is a wrapper function. It first looks for a node containing the old data,
 * deletes this node (if found) and inserts new node with new data.
 *
 * @param t treap
 * @param o old data
 * @param n new data
 */
void treap_modify(struct treap *t, treap_data_t o, treap_data_t n);

/**
 * @brief performs in-order traversal
 *
 * @param t treap
 * @param v visitor function
 * @param priv private data which will be passed to visitor
 */
void treap_traverse_in(struct treap *t, treap_visitor_t v, void *priv);

/**
 * @brief performs level-order traversal
 *
 * @param t treap
 * @param v visitor function
 * @param priv private data which will be passed to visitor
 */
void treap_traverse_level(struct treap *t, treap_visitor_t v, void *priv);

/**
 * @brief returns the maximum height of the tree
 *
 * @param t tree
 *
 * @return height
 */
int treap_max_height_get(struct treap *t);

/**
 * @brief calculates the balance factor of the tree
 *
 * @param t treap
 *
 * @return balance factor (height of right subtree - height of left subtree)
 */
int treap_balance_factor_get(struct treap *t);

/**
 * @brief looks up for node's parent
 *
 * @param t treap
 * @param n node
 *
 * @return parent of the node or NULL if node is root
 */
struct treap_node* treap_parent_find(struct treap *t, struct treap_node *n);

/**
 * @brief looks for the node containing the data
 *
 * @param t treap
 * @param d data
 *
 * @return node containing the data (if found) or NULL (if not found)
 */
struct treap_node* treap_find(struct treap *t, treap_data_t d);

/**
 * @brief returns the minimum value from the treap
 *
 * @param t treap
 *
 * @return node containing minimum value
 */
struct treap_node* treap_min(struct treap *t);

/**
 * @brief returns the maximum value from the treap
 *
 * @param t treap
 *
 * @return node containing maximum value
 */
struct treap_node* treap_max(struct treap *t);

#endif /* end of include guard: TREAP_H_IGVJQTB6 */

